import logging
import os
from typing import Any, List, Optional

from .agregations import adds
from zpy.cli.scaffolding import creator, place_holders
from zpy.cli.scaffolding.boilerplates import basic
from zpy.cli.utils import exist_folder, join_path, if_exist_move, camel_case


def get_last_import(lines):
    index = 0
    for i in range(0, len(lines)):
        if "import" in lines[i] or "from" in lines[i]:
            index = i
    return index + 1


def add_context(name: str, use_case: str, force: bool):
    current_directory = os.getcwd()
    initial_dir = os.getcwd()
    exist, current_directory = if_exist_move(current_directory, 'src')
    exist, current_directory = if_exist_move(current_directory, 'contexts')
    if exist is False and force is False:
        print("Oops!! It seems you are not in a project generated by zpy...")
        print("Can't find the src or contexts directory")
        print("\nIf you want generate the context outside of project use the -f flag when add the context.\n")
        print(f'\nzpy add context -n {name} -uc {use_case} -f\n')
        return
    in_data = [{"value": "", "title": "Project name:"},
               {"value": "", "title": "Project description:"},
               {"value": "", "title": "Author:"},
               {"value": "", "title": "Url Author:"},
               {"value": name.lower(), "title": "Context:"},
               {"value": use_case, "title": "Use Case:"},
               {"value": name.title(), "title": "Repository"},
               {"value": "", "title": "Open Project?:"},
               {"value": "", "title": "Only Content?:"},
               current_directory
               ]
    try:
        creator(current_directory, basic['src']['child']['contexts']['child'], in_data, False, force)
    except Exception as e:
        print(e.args)
        print("Can't create context...")
        return
    try:
        file_writer(join_path(initial_dir, "src/di.py"), adds.get("context_imports").get("code"), in_data)
    except:
        print("Can't update dependencies file... Not Found!")
    print(f"\nContext with name: {name} created successfully!\n")


def add_use_case(name: str, context: str):
    current_directory = os.getcwd()
    initial_dir = os.getcwd()
    exist, current_directory = if_exist_move(current_directory, 'src')
    exist, current_directory = if_exist_move(current_directory, 'contexts')
    exist, current_directory = if_exist_move(current_directory, context)
    exist, current_directory = if_exist_move(current_directory, 'application')
    if exist is False:
        print("Oops!! It seems you are not in a project generated by zpy...")
        print(f"Can't find the src/contexts/{context}/application directory")
        print("\nRemember that you need execute command in the root path of the project.")
        return
    in_data = [{"value": "", "title": "Project name:"},
               {"value": "", "title": "Project description:"},
               {"value": "", "title": "Author:"},
               {"value": "", "title": "Url Author:"},
               {"value": context.lower(), "title": "Context:"},
               {"value": camel_case(name), "title": "Use Case:"},
               {"value": context.title(), "title": "Repository"},
               {"value": "", "title": "Open Project?:"},
               {"value": "", "title": "Only Content?:"},
               current_directory
               ]
    try:
        creator(current_directory,
                basic['src']['child']['contexts']['child']['@context_name']['child']['application']['child'], in_data,
                False, False)
    except Exception as e:
        print(e.args)
        print("Can't create the use case...")
    try:
        file_writer(join_path(initial_dir, "src/di.py"), adds.get("use-cases").get("code"), in_data)
    except:
        print("Can't update dependencies file... Not Found!")
    print(f"\nUseCase with name: {name} created successfully!\n")


def add_config(db):
    print("Add database configuration...")
    di_path = "./src/di.py"
    codes = adds.get("di.py").get("code")
    file_writer(di_path, codes, None)
    print(
        "Execute: \n\n\tvirtualenv venv\n\t./venv/Scripts/activate\n\tpip install -r requirements.txt\n\tConfigure environment variables\n\tpython ./src/local_deploy.py")


def file_writer(di_path: str, codes: dict, data: Optional[List[Any]]):
    print("Searching for dependencies file...")
    try:
        with open(di_path, "r+", encoding="utf-8") as f:
            lines = f.readlines()
            imports_to_add = []
            for line_toad in codes["imports"]:
                to_add = True
                for line in lines:
                    if line.strip("\n\r\t") == line_toad.strip("\n\r\t"):
                        to_add = False
                        break
                if to_add is True:
                    imports_to_add.append(place_holders(line_toad, data))
            final_lines = [place_holders(line, data) for line in codes.get("blocks")]
            index = get_last_import(lines)
            old1 = lines[:index]
            old = lines[index + 1:]
            f.seek(0)
            f.truncate(0)
            f.writelines(old1)
            f.writelines(imports_to_add)
            f.writelines(old)
            f.writelines(final_lines)
        print("Configuration added into di.py...")
    except Exception as e:
        logging.exception(e)



from typing import Optional
import os


def get_env(key: str, default: Optional[str] = None) -> Optional[str]:
    """Get environment value by key

    Parameters:
        key (str): environment variable name
        default (Optional[str]): default value if environment variable not found

    Returns:
        Optional[str]:Returning environment varibale value or default value
    """

    try:
        value: str = os.getenv(key, default)
        return value if value is not None else default
    except Exception as _:
        return default


def get_env_or_throw(key: str) -> str:
    """Get environment value by key

    Parameters:
        key (str): environment variable name
    Raises:
        RuntimeError: Variable name not found
    Returns:
        str:Returning environment varibale value or default value
    """

    value: str = os.environ[key]

    if value is None:
        raise Exception(f"Environment variable '{key}' not found")

    return value



from flask.wrappers import Request
from typing import Optional, List

from zpy.api.http.header_validator import ZHeader, get_headers_or_raise


def flask_get_headers(
    request: Request,
    headers: List[str],
    throw: Optional[bool] = True,
    check_value: Optional[bool] = True,
    include_all: Optional[bool] = False,
    lower_headers: Optional[bool] = True,
) -> ZHeader:
    """
    Check if keys provided as headers are in the headers dict request

    :param request: Request flask
    :param throw : Raise exception if some header not found
    :param check_value: Verify header are in request and value is different of null or empty
    :param include_all: return all headers
    :param headers: headers key to validate
    :param lower_headers: Apply lower case to header name's
    :return: dict with headers
    """
    raw_headers: dict = dict(
        zip(request.headers.keys(lower_headers), request.headers.values())
    )
    return get_headers_or_raise(raw_headers, headers, throw, check_value, include_all)



from typing import Any, Callable, List, Optional
from functools import reduce
import logging
import time
import inspect
from zpy.utils.values import if_null_get
from zpy.app import zapp_context as ctx


def safely_exec(callable_fn: Callable, args=None) -> Optional[Any]:
    """
    Execute provided function in try:except block
    @param callable_fn:
    @param args:
    @return: value returned of execution or none
    """
    if args is None:
        args = []
    try:
        return callable_fn(*args)
    except Exception as e:
        logging.exception(e)
    return None


def safe_exec_wrapper(target: Callable, args=None, kwargs: dict = None,
                      msg: str = None,
                      notifier: Optional[Callable] = None,
                      default_ret=None):
    logg = ctx().logger
    if args is None:
        args = []
    if not kwargs:
        kwargs = {}
    try:
        return target(*args, **kwargs)
    except Exception as e:
        msg = if_null_get(msg, f"An error occurred while try execute: {target.__name__} with: {args} and {kwargs}.")
        logg.exception(msg)
        if notifier:
            notifier(f"Fatal: {msg}\n{str(e)}")
    return default_ret


def exec_if_nt_null(callable_fn: Callable, args: Optional[List[Any]] = None) -> object:
    """
    Execute function if args not null
    """
    if args is None:
        args = []
    for arg in args:
        if arg is None:
            return False
    return callable_fn(*args)


def safely_exec_with(callable_fn: Callable, default_value: Any = None, args=None) -> Optional[Any]:
    """
    Execute provided function in try:except block
    @param default_value:
    @param callable_fn:
    @param args:
    @return: value returned of execution or none
    """
    if args is None:
        args = []
    try:
        return callable_fn(*args)
    except Exception as e:
        logging.exception(e)
    return default_value


def get_class_that_defined_method(meth):
    for cls in inspect.getmro(meth.im_self):
        if meth.__name__ in cls.__dict__:
            return cls
    return None


def timeit(msg: str = None):
    def _timeit_(method):
        def timed(*args, **kw):
            logg = ctx().logger
            ts = time.time()
            result = method(*args, **kw)
            te = time.time()

            method_name = ('{} -> {}'.format(method.__module__, method.__name__)) if not msg else msg
            if 'log_time' in kw:
                name = kw.get('log_name', method.__name__.upper())
                kw['log_time'][name] = int((te - ts) * 1000)
            else:
                logg.info(f"Time Execution: {method_name} :: {(te - ts) * 1000:2.2f} ms.")

            return result

        return timed

    return _timeit_


def fn_composite(*func):
    """
    Function composition
    @param func: functions
    @return: composition
    """

    def compose(f, g):
        return lambda x: f(g(x))

    return reduce(compose, func, lambda x: x)



from typing import Dict, List, Tuple, Union, Any, Optional

from marshmallow.exceptions import ValidationError

from zpy.api.http.errors import BadRequest, ZHttpError
from zpy.utils.objects import ZObjectModel


def parse_request(
        request: dict, model: Union[ZObjectModel, Any], raise_err: bool = True
) -> Tuple[Optional[ZObjectModel], ZHttpError]:
    """
    Parse and validate request according model specification.

    @param request:
    @param model:
    @param raise_err:
    @return:
    """

    model_result: Union[List, ZObjectModel] = None
    errors: Union[List, None, BadRequest] = None
    try:
        if request is None or len(request.items()) == 0:
            error = BadRequest(
                "The request was not provided, validation request error",
                f"Missing fields {model().__missing_fields__} not provided",
            )
            if raise_err:
                raise error
            return None, error
        model_result = model(**request)
    except ValidationError as e:
        model_result = e.valid_data
        # if isinstance(e.messages, Dict):
        #     errors = [e.messages]
        # else:
        #     errors = e.messages
        errors = BadRequest(None, f"{e.messages}")
        if raise_err:
            raise errors
    return model_result, errors



from flask.wrappers import Request, Response
from abc import ABC, abstractmethod
from flask import Flask
from typing import Any

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "upax"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


# Middlewares | Zurck'Z Middlware
# Base middleware for flask
class ZMiddleware(ABC):
    def __init__(self, app: Flask, **kwargs) -> None:
        super().__init__()
        self.app = app
        self.kwargs = kwargs

    @abstractmethod
    def __call__(self, environ: Any, start_response: Any) -> Any:
        return self.app(environ, start_response)


class ParserMiddleWare(ZMiddleware):
    """
    Default middleware for custom access response
    """

    def __init__(self, app: Flask, **kwargs):
        super().__init__(app, **kwargs)
        self.app = app

    def __call__(self, environ, start_response):
        request = Request(environ)
        return self.app(environ, start_response)



# Created by Noé Cruz | Zurckz 22 at 23/04/2022
# See https://www.linkedin.com/in/zurckz
from multiprocessing import Pool, cpu_count, Process, Queue
from typing import Any, Callable, Iterable, List, Optional, Tuple
from zpy.utils.funcs import timeit

_func: Optional[Callable[[Any], Any]] = None


def worker_init(func):
    global _func
    _func = func


def worker(x):
    return _func(x)


def _task_wrapper(idx, queue, task, args):
    queue.put((idx, task(*args)))


def mapp(
        collection: Iterable[Any],
        fn: Callable[[Any], Any],
        chunk_size: Optional[int] = 1,
        args: Optional[Tuple[Any]] = None,
) -> List[Any]:
    """Parallel Collection Processor

    Args:
        collection (Iterable[Any]): Iterable
        fn (Callable[[Any], Any]): Map function
        chunk_size (Optional[int]): chunk size. Default 1
        args: Args
    Returns:
        List[Any]: iterable

    """
    n_cpu = cpu_count()
    if args:
        with Pool(processes=n_cpu) as pool:
            return pool.starmap(fn, [(e,) + args for e in collection], chunk_size)

    with Pool(processes=n_cpu, initializer=worker_init, initargs=(fn,)) as pool:
        return pool.map(worker, collection, chunk_size)


def runp(
        tasks: List[Callable[[Any], Any]], args: Optional[List[Tuple[Any]]] = None
) -> List[Any]:
    """Run tasks in parallel.

    Args:
        tasks (List[Callable[[Any], Any]]): Collection of tasks references
        args (Optional[List[Tuple[Any]]], optional): Args of tasks. Defaults to None.

    Raises:
        ValueError: if the number of args and tasks arent the same

    Returns:
        List[Any]: Ordered Tasks result
    """
    if args is not None and len(tasks) != len(args):
        raise ValueError("Number of args must be equal to number of tasks.")

    queue = Queue()
    processes = [
        Process(
            target=_task_wrapper,
            args=(i, queue, task, () if not args else args[i]),
        )
        for i, task in enumerate(tasks)
    ]
    for process in processes:
        process.start()
    for process in processes:
        process.join()

    return [value[1] for value in sorted([queue.get() for _ in processes])]



from re import sub
import os
from typing import Tuple


def exist_folder(folder: str, path: str) -> bool:
    return os.path.isdir(os.path.join(path, folder))


def join_path(path: str, directory):
    return os.path.join(path, directory)


def if_exist_move(path: str, folder: str) -> Tuple[bool, str]:
    move = exist_folder(folder, path)
    if move:
        current_directory = join_path(path, folder)
        os.chdir(current_directory)
        return True, current_directory
    return False, path


def is_camel_case(s):
    return s != s.lower() and s != s.upper() and "_" not in s


def camel_case(s):
    if is_camel_case(s):
        return s
    s = sub(r"(_|-)+", " ", s).title().replace(" ", "")
    return ''.join([s[0].title(), s[1:]])



from datetime import datetime
from time import strftime, gmtime
from typing import Optional

from dateutil import tz


def second_to_strf(seconds: int, str_format: str = "%H:%M:%S") -> str:
    """Format seconds to format

    Args:
        @param seconds: seconds
        @param str_format: format

    Returns:
        str: seconds in string format

    """
    return strftime(str_format, gmtime(seconds))


def get_tz_info(tz_name: str):
    return tz.gettz(tz_name)


def str_to(str_date: str, date_format: str = '%d/%m/%y %H:%M:%S') -> datetime:
    return datetime.strptime(str_date, date_format)


def apply_tz(date_time: datetime, tz_name: Optional[str] = 'America/Mexico_City') -> datetime:
    tz_info = tz.gettz(tz_name)
    return date_time.astimezone(tz_info)



import logging
from typing import Any, Callable, Tuple, Optional

from .serializers import serialize_object_value
from zpy.api.http.errors import ZHttpError
from zpy.api.http.status_codes import HttpStatus
from zpy.containers import shared_container
from zpy.utils.objects import ZObjectModel
from zpy.utils.values import if_null_get

_useAwsRequestId = False
_custom_request_id_key: Optional[str] = None
_wrapPayloadKey = None
_useStatusFields = False
_custom_status_key: Optional[str] = None
_custom_message_key: Optional[str] = None
_custom_builder: Optional[Callable[[dict, HttpStatus, dict], Tuple[dict, int, dict]]] = None
_custom_error_builder: Optional[Callable[[dict, HttpStatus, dict], Tuple[dict, int, dict]]] = None

_custom_err_code_key: Optional[str] = None
_custom_err_message_key: Optional[str] = None
_custom_err_details_key: Optional[str] = None
_custom_err_meta_key: Optional[str] = None


def setup_error_response(
        custom_code_key: Optional[str] = None,
        custom_message_key: Optional[str] = None,
        custom_details_key: Optional[str] = None,
        custom_metadata_key: Optional[str] = None
):
    global _custom_err_code_key
    _custom_err_code_key = custom_code_key
    global _custom_err_message_key
    _custom_err_message_key = custom_message_key
    global _custom_err_details_key
    _custom_err_details_key = custom_details_key
    global _custom_err_meta_key
    _custom_err_meta_key = custom_metadata_key


def setup_response_builder(
        use_aws_request_id: bool = False,
        wrap_payload_key: Optional[str] = None,
        use_status_fields: bool = False,
        custom_builder: Callable = None,
        custom_error_builder: Callable = None,
        custom_request_id_key: Optional[str] = None,
        custom_status_key: Optional[str] = None,
        custom_message_key: Optional[str] = None
) -> None:
    """
    Configure response builder according you need it
    @param use_aws_request_id: True if you work with aws services and need add aws request id into your response
    @param wrap_payload_key: Pass string key for wrap your payload
    @param use_status_fields: True if you need add 'status' and 'message' fields into response
    @param custom_builder: Function to update or build response according you need it
    @param custom_error_builder: Function to update or build error response according you need it
    @param custom_request_id_key: Custom name for request id
    @param custom_message_key: Custom key name for message prop
    @param custom_status_key: Custom key name for message prop
    @return: None
    """
    global _useAwsRequestId
    _useAwsRequestId = use_aws_request_id
    global _wrapPayloadKey
    _wrapPayloadKey = wrap_payload_key
    global _useStatusFields
    _useStatusFields = use_status_fields
    global _custom_builder
    _custom_builder = custom_builder
    global _custom_error_builder
    _custom_error_builder = custom_error_builder
    global _custom_request_id_key
    _custom_request_id_key = custom_request_id_key
    global _custom_status_key
    _custom_status_key = custom_status_key
    global _custom_message_key
    _custom_message_key = custom_message_key


class SuccessResponse(object):
    def __init__(self, status: HttpStatus = HttpStatus.SUCCESS) -> None:
        self._http_status_ = status

    @property
    def http_status(self):
        return self._http_status_

    @staticmethod
    def empty(status: HttpStatus = HttpStatus.SUCCESS):
        res = ZTResponse()
        res._http_status_ = status
        return res


class ZTResponse(ZObjectModel, SuccessResponse):
    def __init__(self, **kwargs):
        ZObjectModel.__init__(self, use_native_dumps=True, **kwargs)
        SuccessResponse.__init__(self)


def update_error_response(response: dict, status: HttpStatus = HttpStatus.INTERNAL_SERVER_ERROR,
                          headers: Optional[dict] = None) -> Tuple[dict, int, dict]:
    """
    Http error response updater
    @param response: error response content
    @param status: http status
    @param headers: headers
    @return: tuple with response data

    @contact https://www.linkedin.com/in/zurckz
    @author Noé Cruz | Zurck'z 20
    @since 16-05-2020
    """
    http_status_code: int = status.value[0]
    if _useAwsRequestId is True:
        if type(response) is dict and "aws_request_id" in shared_container:
            response.update({f"{if_null_get(_custom_request_id_key, 'requestId')}": shared_container["aws_request_id"]})
    if _custom_error_builder is not None:
        response, code, headers = _custom_error_builder(response, status, headers)
        http_status_code = code
    return response, http_status_code, headers


def update_response(
        payload: dict, status: HttpStatus = HttpStatus.SUCCESS, headers: Optional[dict] = None
) -> Tuple[dict, int, dict]:
    """
    Http response updater
    @param payload:
    @param status:
    @param headers:
    @return: tuple with reponse data

    @contact https://www.linkedin.com/in/zurckz
    @author Noé Cruz | Zurck'z 20
    @since 16-05-2020
    """
    http_status_code = status.value[0]
    if _wrapPayloadKey is not None and _wrapPayloadKey != "":
        payload = {_wrapPayloadKey: payload}
    if _useStatusFields is True:
        if type(payload) is dict:
            _key_m = if_null_get(_custom_message_key, 'message')
            _key_s = if_null_get(_custom_status_key, 'status')
            payload.update({f"{_key_s}": status.value[1], f"{_key_m}": status.value[2]})
    if _useAwsRequestId is True:
        _key = if_null_get(_custom_request_id_key, 'requestId')
        payload.update({f"{_key}": None})
        if type(payload) is dict and "aws_request_id" in shared_container:
            payload.update({f"{_key}": shared_container["aws_request_id"]})
    if _custom_builder is not None:
        payload, code, headers = _custom_builder(payload, status, headers)
        http_status_code = code
    return payload, http_status_code, headers


def response_builder(content_type: Optional[str] = 'application/json', headers: Optional[dict] = None):
    """
    HTTP Response builder, build response from data provided
    @param content_type: Response content type. Default: application/json
    @param headers: Extra headers
    @return: None

    @contact https://www.linkedin.com/in/zurckz
    @author Noé Cruz | Zurck'z 20
    @since 16-05-2020
    """
    extra_headers = {'Content-Type': content_type}

    def z_inner_builder(invoker: Callable):
        def wrapper_handler(*args, **kwargs):
            try:
                payload: Any = invoker(*args, **kwargs)

                if issubclass(payload.__class__, ZObjectModel) and issubclass(
                        payload.__class__, SuccessResponse
                ):
                    return update_response(payload.sdump(), payload.http_status)

                if issubclass(payload.__class__, ZObjectModel):
                    return update_response(payload.sdump())

                if issubclass(payload.__class__, SuccessResponse):
                    return update_response(
                        serialize_object_value(payload), payload.http_status
                    )
                return update_response(serialize_object_value(payload), headers=extra_headers)
            except ZHttpError as e:
                logging.exception("[ZCE] - An error was generated when processing request.")
                http_status: HttpStatus = e.status
                if e.reason is not None:
                    e.details = [e.reason] + if_null_get(e.details, [])
                err_response = {
                    f"{if_null_get(_custom_err_message_key, 'message')}": if_null_get(e.message, http_status.value[2]),
                    f"{if_null_get(_custom_err_code_key, 'code')}": http_status.value[1],
                    f"{if_null_get(_custom_err_code_key, 'details')}": e.details,
                }
                if e.metadata is not None:
                    err_response[f'{if_null_get(_custom_err_meta_key, "metadata")}'] = e.metadata
                return update_error_response(
                    err_response,
                    http_status
                )
            except Exception as e:
                logging.exception(
                    "[UCE] - An unexpected error was generated when processing request.", exc_info=e
                )
                code: HttpStatus = HttpStatus.INTERNAL_SERVER_ERROR
                return update_error_response(
                    {f"{if_null_get(_custom_err_message_key, 'message')}": code.value[2],
                     f"{if_null_get(_custom_err_code_key, 'code')}": code.value[1],
                     f"{if_null_get(_custom_err_code_key, 'details')}": None}, code
                )

        wrapper_handler.__name__ = invoker.__name__
        return wrapper_handler

    return z_inner_builder



import logging
from typing import List, Dict
from .boilerplates import basic
from .placeholders import place_holders
import os

from .ignore import ignore


def scaffolding_dispatcher(in_data: List[Dict[str, str]], verbose=False, click=None):
    basic_project(in_data, verbose)


def basic_project(in_data: List[Dict[str, str]], verbose=False):
    try:
        current_directory = os.getcwd()
        name = in_data[0]["value"]
        only_content = in_data[8]["value"]
        force = True
        # Path
        path = current_directory
        if only_content is False:
            force = False
            path = os.path.join(current_directory, name)
            if os.path.isdir(path) is True:
                print(f"Project {name} already exist!")
                return
                # mode
            mode = 0o666
            os.mkdir(path, mode)
            if verbose is True:
                print("Directory '% s' created" % path)
            os.chdir(path)

        if verbose is True:
            print("Working in '% s'" % path)
        in_data[9] = path
        creator(path, basic, in_data, verbose, force)

        if in_data[7]["value"] is True:
            try:
                os.system(f"code {path}")
            except Exception as ex:
                ...
    except Exception as e:
        print("An critical error occurred while try to generate project...")
        if verbose is True:
            logging.exception(e)


def create_dir_and_move(current_directory, name, verbose, force) -> str:
    path = os.path.join(current_directory, name)
    if os.path.isdir(path) is True:
        print(f"Project {name} already exist!")
        return
    mode = 0o666
    os.mkdir(path, mode)
    if verbose is True:
        print("Directory '% s' created" % path)
    os.chdir(path)
    return path


def creator(c_dir: str, context: dict, info: List, verbose: bool, force: bool):
    for k in context.keys():
        os.chdir(c_dir)
        c_file = context.get(k)
        if c_file["type"] == 'file':
            with open(place_holders(c_file["name"], info), "w+", encoding="utf-8") as f:
                f.write(place_holders(c_file["content"], info))
        elif c_file["type"] == "raw-file":
            with open(place_holders(c_file["name"], info), "w+", encoding="utf-8") as f:
                f.write(ignore)
        elif c_file["type"] == "dir":
            creator(create_dir_and_move(c_dir, place_holders(c_file["name"], info), verbose, force), c_file["child"],
                    info, verbose, force)



from typing import Any
import hashlib
import base64
import hmac


def if_null_get(value: Any, default: Any) -> Any:
    """
    Evaluate the given value, if is None return default argument
    otherwise given value
    """
    return default if value is None else value


def get_hmac(key: str, msg: str) -> str:
    """Hash message authentication code generator

    Args:
        key (str): [The starting key for the hash]
        msg (str): [Initial input for the hash, or None.]

    Returns:
        str: [Hash]
    """
    hash_value: bytes = hmac.new(
        str(key).encode("utf-8"), msg=str(msg).encode("utf-8"), digestmod=hashlib.sha256
    ).digest()
    return base64.b64encode(hash_value).decode()


def get_hash(value: str, charset: str = "utf-8") -> str:
    """Hash Generator

    Args:
        value (str): value

    Returns:
        str: sha256 hash
        @param value:
        @param charset:
    """
    return hashlib.sha256(value.encode(charset)).hexdigest()


def is_built_type(value: Any) -> bool:
    """Check if type of value provides is built in Python

    Args:
        value (Any): value

    Returns:
        [bool]: true if type of value is built in Python
    """
    return type(value) in (
        bytearray,
        str,
        list,
        tuple,
        range,
        dict,
        int,
        complex,
        set,
        bytes,
        bool,
        float,
    )



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")
S = TypeVar("S")


class UseCase(ABC, Generic[T, S]):
    @abstractmethod
    def execute(self, input: T, *args, **kwargs) -> S:
        """Execute use case"""
        pass



import logging
import traceback
from abc import ABC, abstractmethod
from enum import Enum
from io import StringIO
from typing import Union, Any, Optional, Callable


class ZLFormat(Enum):
    M = "%(message)s"
    NM = "%(name)s %(message)s"
    LNM = "%(name)s %(levelname)s %(message)s"
    TM = "%(asctime)s %(message)s"
    LM = "%(levelname)s - %(message)s"
    TLM = "%(asctime)s - %(levelname)s - %(message)s"
    TNLM = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"


def setup_global_logger(log_format: str = ZLFormat.LM.value, level: int = logging.INFO, reset: Optional[bool] = False):
    if reset is True:
        if logging.getLogger().hasHandlers() is True:
            logger = logging.getLogger()
            for h in logger.handlers:
                logger.removeHandler(h)
    logging.basicConfig(format=log_format, level=level)


class ZLogger(ABC):
    """
    Simple Logger
    """

    @classmethod
    @abstractmethod
    def create(cls, name: str):
        ...

    @abstractmethod
    def raw(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        ...

    @abstractmethod
    def info(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        ...

    @abstractmethod
    def warn(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        ...

    @abstractmethod
    def err(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        ...

    @abstractmethod
    def ex(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        ...

    @abstractmethod
    def release(self, name: Optional[str] = None) -> None:
        ...

    @abstractmethod
    def close(self) -> None:
        ...


class Shippable:

    def __init__(self):
        self.senders = []

    def emit_if(self, value: Any, predicate):
        if predicate:
            self.emit(value)

    def emit(self, value: Any):
        for sender in self.senders:
            sender(value)

    def add_sender(self, sender: Callable[[Any], None]) -> None:
        self.senders.append(sender)


class ZStreamLogger(ZLogger, Shippable):
    """
    Stream Logger
    """

    def __init__(self, name: str, level: int, log_format: Union[str, ZLFormat]) -> None:
        super().__init__()
        self._level = level
        self._name = name
        self._log_format = format if isinstance(format, str) else log_format.value
        self._formatter = logging.Formatter(self._log_format)
        self._logger = None
        self._stream = None
        self._handler = None
        self.release(name)

    @classmethod
    def create(cls, name: str):
        return cls(name, logging.INFO, ZLFormat.TNLM)

    def raw(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        self._logger.log(self._level, value, **kwargs)

    def info(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        self._logger.info(value, **kwargs)

    def warn(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        self._logger.warning(value, **kwargs)

    def err(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        self._logger.error(value, *kwargs)

    def ex(self, value: Any, shippable: Optional[bool] = False, **kwargs) -> None:
        self._logger.exception(msg=value, **kwargs)

    def release(self, name: Optional[str] = None) -> None:
        self.close()
        self._logger = logging.getLogger(self._name if not name else name)
        for h in self._logger.handlers:
            self._logger.removeHandler(h)
        self._logger.setLevel(self._level)
        self._stream = StringIO()
        self._handler = logging.StreamHandler(stream=self._stream)
        self._handler.setLevel(self._level)
        self._handler.setFormatter(self._formatter)
        self._logger.addHandler(self._handler)
        self._logger.propagate = False

    def close(self) -> None:
        if self._handler:
            self._handler.flush()
            self._handler.close()
        if self._logger:
            self._logger.removeHandler(self._handler)
        if self._stream:
            self._stream.close()
            self._stream = None

    def str_value(self) -> str:
        return self._stream.getvalue()

    def send(self):
        if self._stream:
            self.emit(self._stream.getvalue())


class zL(ZLogger, Shippable):
    """
    Logger Wrapper
    """

    def __init__(self, name: str, level: int, log_format: Union[str, ZLFormat]) -> None:
        super().__init__()
        self._level = level
        self._name = name
        self._log_format = format if isinstance(format, str) else log_format.value
        self._formatter = logging.Formatter(self._log_format)
        self._logger = None
        self._handler = None
        self.release(name)

    def release(self, name: Optional[str] = None):
        self._logger = logging.getLogger(self._name if not name else name)
        for h in self._logger.handlers:
            self._logger.removeHandler(h)
        self._logger.setLevel(self._level)
        self._handler = logging.StreamHandler()
        self._handler.setLevel(self._level)
        self._handler.setFormatter(self._formatter)
        self._logger.addHandler(self._handler)
        self._logger.propagate = False

    @classmethod
    def create(cls, name: str):
        return cls(name, logging.INFO, ZLFormat.TNLM)

    @classmethod
    def create_for_cloud(cls, name: str):
        return cls(name, logging.INFO, ZLFormat.LM)

    def raw(self, value: Any, shippable: bool = False, *args, **kwargs):
        self.emit_if(value, shippable)
        self._logger.log(self._level, value, *args, **kwargs)

    def info(self, value: Any, shippable: bool = False, **kwargs) -> None:
        self.emit_if(value, shippable)
        self._logger.info(value, **kwargs)

    def warn(self, value: Any, shippable: bool = False, **kwargs) -> None:
        self.emit_if(value, shippable)
        self._logger.warning(value, **kwargs)

    def err(self, value: Any, shippable: bool = False, **kwargs) -> None:
        self.emit_if(value, shippable)
        self.emit_if(traceback.format_exc(), shippable)
        self._logger.error(value, *kwargs)

    def exception(self, value: Any, shippable: bool = False, **kwargs) -> None:
        self.emit_if(traceback.format_exc(), shippable)
        self.emit_if(traceback.format_exc(), shippable)
        self._logger.exception(msg=value, **kwargs)

    def close(self):
        if self._handler:
            self._handler.flush()
            self._handler.close()
        if self._logger:
            self._logger.removeHandler(self._handler)

    @staticmethod
    def w(msg: object, *args):
        logging.warning(msg=msg, *args)

    @staticmethod
    def i(msg: object, *args):
        """Information Level Log

        Args:
            msg (object): value
        """
        logging.info(msg=msg, *args)

    @staticmethod
    def e(msg: object, exc_info=None, *args):
        """Error Level Log

        Args:
            @param msg:
            @param exc_info:
        """
        logging.error(msg=msg, exc_info=exc_info, *args)

    @staticmethod
    def ex(msg: object, **kwargs):
        """Exception Level Log

        Args:
            @param msg:
        """
        logging.exception(msg=msg, **kwargs)

    @staticmethod
    def d(msg: object, *args):
        """Debug Level Log

        Args:
            msg (object): value
        """
        logging.debug(msg=msg, *args)



# Created by Noé Cruz | Zurckz 22 at 15/04/2022
# See https://www.linkedin.com/in/zurckz
from typing import Any, Callable, Optional
from zpy.app import zapp_context as ctx
from zpy.utils.funcs import safe_exec_wrapper
from zpy.utils.values import if_null_get


def get_sqs_url_from(sqs_client: Any, sqs_name: str) -> str:
    """
    Retrieve sqs url from name
    @param sqs_client:
    @param sqs_name:
    @return: url
    """
    return sqs_client.get_queue_url(
        QueueName=sqs_name,
    )['QueueUrl']


def send_secure_message_to_sqs(sqs_client: Any, payload: str, origin: str, attributes=None, sqs_name: str = None,
                               sqs_url: str = None, delay: int = 0,
                               error_notifier: Optional[Callable[[str], None]] = None) -> Optional[dict]:
    """
    Helper declarative function to send sqs message


    @param sqs_client: SQS Client built from boto3
    @param payload: Message to send
    @param attributes: Custom attributes to send
    @param origin: Origin of emitted message
    @param sqs_name: SQS Name
    @param sqs_url: SQS Url
    @param delay: Delay seconds
    @param error_notifier: Error reporting function

    @note If sqs name is null, sqs url is required or if sqs_url is null, sqs_name is required

    @return: sqs response
    """
    return safe_exec_wrapper(
        target=send_message_to_sqs,
        args=[sqs_client, payload, origin, attributes, sqs_name, sqs_url, delay],
        kwargs=None,
        msg=f"An error occurred while try to send message to: {if_null_get(sqs_name, sqs_url)}",
        notifier=error_notifier,
        default_ret=None
    )


def send_message_to_sqs(sqs_client: Any, payload: str, origin: str, attributes=None, sqs_name: str = None,
                        sqs_url: str = None, delay: int = 0) -> dict:
    """
    Helper declarative function to send sqs message

    @param sqs_client: SQS Client built from boto3
    @param payload: Message to send
    @param attributes: Custom attributes to send
    @param origin: Origin of emitted message
    @param sqs_name: SQS Name
    @param sqs_url: SQS Url
    @param delay: Delay seconds

    @note If sqs name is null, sqs url is required or if sqs_url is null, sqs_name is required

    @return: sqs response
    """

    if not sqs_name and not sqs_url:
        raise ValueError("SQS Name or SQS Url is required")

    if attributes is None:
        attributes = {}

    if 'ZOrigin' not in attributes:
        attributes['ZOrigin'] = {
            'DataType': 'String',
            'StringValue': origin if origin else 'Unknown'
        }

    return sqs_client.send_message(
        QueueUrl=get_sqs_url_from(sqs_client, sqs_name) if sqs_name else sqs_url,
        DelaySeconds=delay,
        MessageAttributes=attributes,
        MessageBody=payload
    )


def send_wake_up_signal(sqs_client: Any, metadata: str, sqs_name: str = None, sqs_url: str = None,
                        origin: str = "Unknown"):
    logg = ctx().logger
    destination = if_null_get(sqs_name, sqs_url)
    try:
        response = send_message_to_sqs(
            sqs_client=sqs_client,
            payload=metadata,
            origin=origin,
            attributes={
                'ZType': {
                    'DataType': 'String',
                    'StringValue': 'WakeUpSignal'
                }
            },
            sqs_name=sqs_name,
            sqs_url=sqs_url)
        logg.info(f"Wake up signal was sent successfully to {destination}, id: {response['MessageId']}")
    except Exception as e:
        logg.err(f"An error occurred while try to send Wake up signal to: {destination}", exc_info=e)



from datetime import datetime
import re
from typing import List, Optional


def to_snake(text: str) -> str:
    return re.sub(r'(?<!^)(?=[A-Z])', '_', text).lower()


_places_holders = [
    {
        "place": "@project_name",
        "builder": lambda x: x[0]["value"]
    },
    {
        "place": "@project_description",
        "builder": lambda x: x[1]["value"]
    },
    {
        "place": "@repository_name",
        "builder": lambda x: f"{x[6]['value'].title()}Repository"
    },
    {
        "place": "@repository_impl",
        "builder": lambda x: f"Awesome{x[6]['value'].title()}Repository"
    },
    {
        "place": "@repository_action",
        "builder": lambda x: f"{to_snake(x[5]['value']).lower()}"
    },
    {
        "place": "@repository_file",
        "builder": lambda x: f"{x[6]['value'].lower()}_repository"
    },
    {
        "place": "@usecase_file",
        "ex": "GetUser",
        "builder": lambda x: f"{to_snake(x[5]['value'])}"
    },
    {
        "place": "@usecase_file_test",
        "ex": "GetUser",
        "builder": lambda x: f"{to_snake(x[5]['value'])}_test"
    },
    {
        "place": "@usecase_class",
        "ex": "GetUser",
        "builder": lambda x: f"{x[5]['value']}"
    },
    {
        "place": "@usecase_var",
        "ex": "GetUser",
        "builder": lambda x: f"{to_snake(x[5]['value'])}_uc"
    },
    {
        "place": "@developer",
        "builder": lambda x: f"{x[2]['value']}"
    },
    {
        "place": "@dev_url",
        "builder": lambda x: f"{x[3]['value']}"
    },
    {
        "place": "@context_name",
        "builder": lambda x: f"{x[4]['value']}"
    },
    {
        "place": "@date",
        "builder": lambda x: datetime.today().strftime('%Y-%m-%d')
    },
    {
        "place": "@base_path",
        "builder": lambda x: str(x[9]).replace("\\", "\\\\")
    },
]


def place_holders(raw: str, data: Optional[List]) -> str:
    if data is None:
        return raw
    for p in _places_holders:
        try:
            raw = raw.replace(p["place"], p["builder"](data))
        except:
            ...
    return raw



from enum import Enum
from typing import Any, List, Optional, Tuple, Callable, Dict
from requests import post, Response, get, patch, delete, put
from zpy.logger import zL, ZLogger
from datetime import timedelta


class HLevels(Enum):
    URL = 1
    HEADERS = 2
    PARAMS = 3
    PAYLOAD = 4

    @staticmethod
    def all():
        return [HLevels.URL, HLevels.HEADERS, HLevels.PARAMS, HLevels.PAYLOAD]

    @staticmethod
    def url_and_body():
        return [HLevels.URL, HLevels.PAYLOAD]


def zhl_write(text: str, value: Any, logger: ZLogger = None):
    def wrapper():
        if value is not None:
            if not logger:
                zL.i(f"{text} {value}")
                return
            logger.info(f"{text} {value}")

    return wrapper


class ZHttp:
    """HTTP Wrapper


    Raises:
        ValueError: [description]
    """

    global_options: dict = {"BASE_URL": None, "LOG_OPTIONS": None}

    @staticmethod
    def setup(base: str, log_options: List[HLevels] = None) -> None:
        ZHttp.global_options["BASE_URL"] = base
        ZHttp.global_options["LOG_OPTIONS"] = log_options

    @staticmethod
    def __prepare_request__(
            url, path, params, headers, data, json, log_options, logger: ZLogger = None
    ) -> Tuple[str, List[HLevels]]:
        final_url: str = ZHttp.global_options["BASE_URL"]
        if url is not None and path is not None:
            final_url = f"{url}{path}"
        if url is not None and path is None:
            final_url = url
        if url is None and path is not None:
            final_url = f"{final_url}{path}"

        if final_url is None:
            raise ValueError("URL not configured!")

        real_log_options = ZHttp.global_options["LOG_OPTIONS"]

        try:
            data_to_log: Dict[HLevels, Callable] = {
                HLevels.URL: zhl_write(f"Start HTTP [POST] -", final_url, logger),
                HLevels.PARAMS: zhl_write(f"Params:", params, logger),
                HLevels.HEADERS: zhl_write(f"Headers:", headers, logger),
                HLevels.PAYLOAD: zhl_write(f"Body:", data if json is None else json, logger)
            }

            if log_options is not None and len(log_options) > 0:
                real_log_options = log_options

            if not real_log_options:
                return final_url, real_log_options

            for i in real_log_options:
                writer = data_to_log.get(i, None)
                if writer:
                    writer()
        except Exception as e:
            ZHttp.__log_exception__(logger, "An error occurred while logging http request.")
        return final_url, real_log_options

    @staticmethod
    def __log_exception__(logger: ZLogger = None, msg: str = "The http request failed..."):
        if not logger:
            zL.ex(msg)
            return
        logger.info(msg)

    @staticmethod
    def __logging_response__(
            result: Response, final_url, log_response_headers, real_log_options, logger: ZLogger = None
    ):
        if not real_log_options:
            return
        parsed = result.json()
        elapsed_time = str(timedelta(seconds=result.elapsed.total_seconds()))

        response_logs: Dict[HLevels, Callable] = {
            HLevels.URL: zhl_write(f"End HTTP [POST] - {elapsed_time} - {result.status_code}", final_url, logger),
            HLevels.HEADERS: zhl_write(
                f"Response Headers:",
                dict(zip(result.headers.keys(), result.headers.values()))
                if log_response_headers is True
                else None,
                logger
            ),
            HLevels.PAYLOAD: zhl_write(f"Response Body:", parsed, logger)
        }
        for i in real_log_options:
            writer = response_logs.get(i, None)
            if writer:
                writer()

    @staticmethod
    def get(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_options: List[HLevels] = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
            logger: ZLogger = None
    ):
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_options, logger
        )
        try:
            result: Response = get(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level is not None:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level, logger
                )
            return result
        except Exception as e:
            ZHttp.__log_exception__(logger)
            if control_failure is False:
                raise e
        return None

    @staticmethod
    def post(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_options: List[HLevels] = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
            logger: ZLogger = None
    ) -> Optional[Response]:
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_options, logger
        )
        try:
            result: Response = post(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level is not None:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level, logger
                )
            return result
        except Exception as e:
            ZHttp.__log_exception__(logger)
            if control_failure is False:
                raise e
        return None

    @staticmethod
    def put(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_options: List[HLevels] = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
            logger: ZLogger = None
    ) -> Optional[Response]:
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_options, logger
        )
        try:
            result: Response = put(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level is not None:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level, logger
                )
            return result
        except Exception as e:
            ZHttp.__log_exception__(logger)
            if control_failure is False:
                raise e
        return None

    @staticmethod
    def patch(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_options: List[HLevels] = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
            logger: ZLogger = None
    ) -> Optional[Response]:
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_options, logger
        )
        try:
            result: Response = patch(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level is not None:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level, logger
                )
            return result
        except Exception as e:
            ZHttp.__log_exception__(logger)
            if control_failure is False:
                raise e
        return None

    @staticmethod
    def delete(
            url: Optional[str] = None,
            path: Optional[str] = None,
            params: Any = None,
            data: Any = None,
            headers: Any = None,
            cookies: Any = None,
            files: Any = None,
            auth: Any = None,
            timeout: Any = None,
            allow_redirects: bool = None,
            proxies: Any = None,
            hooks: Any = None,
            stream: Any = None,
            verify: Any = None,
            cert: Any = None,
            json: Any = None,
            log_options: List[HLevels] = None,
            log_response_headers: bool = False,
            control_failure: bool = False,
            logger: ZLogger = None
    ) -> Optional[Response]:
        final_url, real_log_level = ZHttp.__prepare_request__(
            url, path, params, headers, data, json, log_options, logger
        )
        try:
            result: Response = delete(
                url=final_url,
                json=json,
                data=data,
                params=params,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                hooks=hooks,
                stream=stream,
                verify=verify,
                cert=cert,
            )
            if real_log_level is not None:
                ZHttp.__logging_response__(
                    result, final_url, log_response_headers, real_log_level, logger
                )
            return result
        except Exception as e:
            ZHttp.__log_exception__(logger)
            if control_failure is False:
                raise e
        return None


class ZHttpResponse(object):
    """
    Wrapper
    """

    @classmethod
    def of(cls, response: Response):
        return cls(response)

    def __init__(self, response: Response):
        self.raw = response

    def json(self) -> dict:
        return self.raw.json()

    def status_is(self, status: int) -> bool:
        return False if self.raw is None else self.raw.status_code == status

    def status_is_and(self, status: int, action: Callable[[Response], Optional[Any]]) -> Optional[Any]:
        """

        @param status:
        @param action:
        @return:
        """
        if self.raw is not None and self.raw.status_code == status:
            return action(self.raw)
        return None

    def is_ok(self) -> bool:
        """

        @return: http status response is success
        """
        return False if self.raw is None else self.raw.ok



from typing import Any, List, Callable, Optional


def find(any_list: List[Any], x_filter: Callable) -> Optional[object]:
    """
    Find item in provided list according filter
    @param any_list: List of any values
    @param x_filter: Filter to apply the search
    @return: element that filter return true otherwise None
    """
    for x in any_list:
        if x_filter(x):
            return x
    return None



# Created by Noé Cruz | Zurckz 22 at 20/03/2022
# See https://www.linkedin.com/in/zurckz

import json
import os
import sys
from pathlib import Path
from typing import Callable, List


def read_event(name: str = None, event_file_path: str = None):
    event_path = f"{os.getcwd()}\\events\\{name}" if name else event_file_path
    if not event_path:
        raise ValueError("Path file or name is required...")
    with open(event_path, 'r') as sm:
        return json.loads(sm.read())


def add_source_to_path(src_dir: str = None, replacement: str = None):
    current_dir = src_dir
    if not current_dir:
        current_dir = os.getcwd()
        current_dir = current_dir.replace('\\tests' if not replacement else replacement, '\\src')
    sys.path.append(current_dir)
    path = Path(current_dir)
    sys.path.append(str(path.parent.absolute()))


def get_files_in(directory: str, ext, absolute=False) -> List[str]:
    path_files = []
    parent = os.getcwd()
    for root, dirs, files in os.walk(directory):
        for filename in files:
            if filename.endswith(ext):
                if absolute:
                    path_files.append(os.path.join(parent, root, filename))
                else:
                    path_files.append(os.path.join(root, filename))
    return path_files


def file_content_updater(
        file_path: str,
        mutator: Callable[[str], str] = None,
        find: str = None,
        replaced: str = None,
) -> None:
    """
    Content File Updater
    @param file_path:
    @param mutator:
    @param find:
    @param replaced:
    @return:
    """
    with open(file_path, "r+b") as f:
        content = f.readlines()
        for i, line in enumerate(content):
            line = line.decode("utf-8")
            if mutator:
                new: str = mutator(line)
                content[i] = new.encode("utf-8")
            else:
                if line and find in line:
                    print(f"Found in {file_path} line: {i}... Removed...")
                    content[i] = line.replace(find, replaced).encode("utf-8")
        f.seek(0)
        f.truncate()
        f.writelines(content)



from enum import Enum

class HttpStatus(Enum):
    """
    Common HTTP status codes

    CODE | SHORT DESCRIPTION | STATUS DETAILS
    """

    SUCCESS                = (200, "SUCCEEDED", "The request has succeeded")
    CREATED                = (201, "CREATED","The request has been fulfilled and resulted in a new resource being created.")
    ACCEPTED               = (202, "ACCEPTED","The request has been accepted for processing, but the processing has not been completed.")
    NO_CONTENT             = (204, "NO CONTENT","The request has been completed successfully but your response has no content, although the headers can be useful.",)
    PARTIAL_CONTENT        = (206, "PARTIAL CONTENT", "Partial content")

    BAD_REQUEST            = (400, "BAD REQUEST","The request could not be understood by the server due to malformed syntax. The client SHOULD NOT repeat the request without modifications.",)
    UNAUTHORIZED           = (401, "UNAUTHORIZED", "The request requires user authentication.")
    FORBIDDEN              = (403, "FORBIDDEN","The server understood the request, but is refusing to fulfill it.")
    NOT_FOUND              = (404, "NOT FOUND","The server has not found anything matching the Request-URI.",)
    METHOD_NOT_ALLOWED     = (405, "METHOD NOT ALLOWED","The method specified in the Request-Line is not allowed for the resource identified by the Request-URI.",)
    CONTENT_NOT_ACCEPTABLE = (406, "METHOD NOT ACCEPTABLE","The resource identified by the request is only capable of generating response entities which have content characteristics not acceptable according to the accept headers sent in the request.",)
    REQUEST_TIMEOUT        = (408, "REQUEST TIMEOUT", "Time out")
    PRE_CONDITION_FAILED   = (412, "PRECONDITION FAILED","The client has indicated preconditions in their headers which the server does not meet.",)
    UNSUPPORTED_MEDIA_TYPE = (415, "UNSUPPORTED MEDIA TYPE","The multimedia format of the requested data is not supported by the server, therefore the server rejects the request.",)
    IM_A_TEAPOT            = (418, "IM A TEAPOT","The server refuses to try to make coffee with a kettle.",)
    CONFLICT               = (409, "CONFLICT", "The server found conflict with request supplied.")
    UNPROCESSABLE          = (422, "UNPROCESSABLE ENTITY","The process could not be completed due to a semantics error.",)
    LOCKED                 = (423, "LOCKED","The source or destination resource of a method is locked.",)

    INTERNAL_SERVER_ERROR  = (500, "INTERNAL SERVER ERROR","The server encountered an unexpected condition which prevented it from fulfilling the request.",)
    NOT_IMPLEMENTED         = (501, "NOT IMPLEMENTED","The server does not support the functionality required to fulfill the request",)
    SERVICE_UNAVAIBLE      = (503, "SERVICE UNAVAIBLE","The server is currently unable to handle the request due to a temporary overloading or maintenance of the server.",)
    GATEWAY_TIMEOUT        = (503, "GATEWAY TIMEOUT", "Timeout")
    LOOP_DETECTED          = (508, "LOOP DETECTED","The server encountered an infinite loop while processing the request. ",)



import json
import typing
from abc import ABC, abstractmethod
from copy import copy
from datetime import datetime, date
from decimal import Decimal
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, NamedTuple

from marshmallow import utils
from marshmallow.fields import Field
from marshmallow_objects import models

from zpy.api.http.errors import BadRequest, ZHttpError
from zpy.utils.funcs import safely_exec
from zpy.utils.values import if_null_get

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"


class IntID(NamedTuple):
    """
    Value object to represent an integer id

    Must be an integer value.
    Must be greater than or equal to 0
    """
    value: int

    @staticmethod
    def of(value: int, entity_name: str = '', validator: Optional[Callable[[int], int]] = None):
        """
        Create new id value object
        @param value:
        @param entity_name: Entity to which the id belongs
        @param validator: Custom validator
        @return: IntID instance
        """
        if validator is not None:
            return IntID(validator(value))
        value = safely_exec(lambda x: int(value), [value])
        if isinstance(value, int) is False:
            raise BadRequest(f"Invalid value for {entity_name} identifier provided.",
                             "The value of id must be integer.",
                             meta={"value": value})
        if value < 0:
            raise BadRequest(f"Invalid value for {entity_name} identifier provided.",
                             "The value of id must be greater than or equal to 0.",
                             meta={"value": value})

        return IntID(value)

    @property
    def val(self):
        """
        id value
        @return: int value
        """
        return self.value


class IntValue(NamedTuple):
    """
    Value object to represent an integer id

    Must be an integer value.
    Must be greater than or equal to 0
    """
    value: int

    @staticmethod
    def of(value: int, entity_name: str = '', validator: Optional[Callable[[int], int]] = None,
           allow_negative: bool = True):
        """
        Create new id value object
        @param allow_negative:
        @param value:
        @param entity_name: Entity to which the id belongs
        @param validator: Custom validator
        @return: IntID instance
        """
        if validator is not None:
            return IntValue(validator(value))
        value = safely_exec(lambda x: int(value), [value])
        if isinstance(value, int) is False:
            raise BadRequest(f"Invalid value for '{entity_name}' provided.",
                             "The data type of value provided must be integer.",
                             meta={"value": value})
        if not allow_negative and value < 0:
            raise BadRequest(f"Invalid provided value for '{entity_name}'.",
                             "The value must be greater than or equal to 0.",
                             meta={"value": value})

        return IntValue(value)

    @property
    def val(self):
        """
        id value
        @return: int value
        """
        return self.value


def default_remove():
    return [
        "_http_status_",
        "__dump_lock__",
        "__schema__",
        "__missing_fields__",
        "__setattr_func__",
        "_ZObjectModel__remove_keys",
        "_ZObjectModel__update_items",
        "__use_native_dumps__",
        "dump_mode"
    ]


class ZObjectModel(models.Model):
    """
    Zurckz Model
    """

    def __init__(
            self,
            exclude: Optional[List[str]] = None,
            include: Optional[Dict[Any, Any]] = None,
            context=None,
            partial=None,
            use_native_dumps=False,
            **kwargs
    ):
        super().__init__(context=context, partial=partial, **kwargs)
        self.__remove_keys = default_remove() + if_null_get(exclude, [])
        self.__update_items = if_null_get(include, {})
        self.__use_native_dumps__ = use_native_dumps

    def __str__(self):
        """
        Dump nested models by own properties
        """
        data = copy(self.__dict__)
        if self.__update_items is not None:
            data.update(self.__update_items)
        [data.pop(k, None) for k in self.__remove_keys]
        for k in data.keys():
            if isinstance(data[k], models.Model):
                data[k] = json.loads(str(data[k]))
            elif isinstance(data[k], list):
                data[k] = [json.loads(str(it)) for it in data[k]]
            elif isinstance(data[k], datetime) or isinstance(data[k], date):
                data[k] = str(data[k])
            elif isinstance(data[k], Decimal):
                data[k] = float(data[k])
        return json.dumps(data)

    def nat_dump(
            self,
            exclude_keys: Optional[List[str]] = None,
            include: Optional[Dict[Any, Any]] = None,
            mutator: Optional[Callable[[Dict], Dict]] = None,
            map_args: Optional[List[Any]] = None,
            store_ex: bool = False,
            store_in: bool = False
    ):
        """
        Dump object using native strategy
        @param exclude_keys:
        @param include:
        @param mutator:
        @param map_args:
        @param store_ex:
        @param store_in:
        @return:
        """
        return self.sdump(exclude_keys, include, mutator, map_args, store_ex, store_in, True)

    def sdump(
            self,
            exclude_keys: Optional[List[str]] = None,
            include: Optional[Dict[Any, Any]] = None,
            mutator: Optional[Callable[[Dict], Dict]] = None,
            map_args: Optional[List[Any]] = None,
            store_ex: bool = False,
            store_in: bool = False,
            use_native_dumps=False
    ):
        """
        Model dump to json safely, checking the exclude key list

        Use this function instead of zdump.

        Parameters:
        -----------

        exclude_keys: List[str], Optional,
            List of string keys of exlude in dump process
        include: Dict[Any,Any], Optional,
            Object to include in model object after exclude process before of dump process
        mutator: Callable, Optional
            Callable function to tranform object after exclude and include process
        map_args: List[Any], Optional
            Argument list to passed to map callable function
        store_ex: bool, optional
            Indicate that the exclude key added to global model exclude key array
        store_in: bool, optional
            Indicate that the include object added to global model object
        """
        data = copy(self.__dict__)

        if map_args is None:
            map_args = []

        native = use_native_dumps if use_native_dumps is True else self.__use_native_dumps__
        if native is True:
            with self.__dump_mode_on__():
                data = self.__schema__.dump(self)
        temp_exclude = copy(self.__remove_keys)
        if exclude_keys is not None:
            temp_exclude = self.__remove_keys + exclude_keys
            if store_ex:
                self.__remove_keys = self.__remove_keys + exclude_keys
        [data.pop(k, None) for k in temp_exclude]
        temp_include = copy(self.__update_items)
        if include is not None:
            temp_include.update(include)
            data.update(temp_include)
            if store_in:
                self.__update_items.update(include)
        else:
            if temp_include is not None:
                data.update(temp_include)
        if mutator is not None:
            data = mutator(data, *map_args)
        if native is True:
            return data
        # TODO Verify this process when native is False
        for k in data.keys():
            if isinstance(data[k], models.Model):
                data[k] = json.loads(str(data[k]))
            elif isinstance(data[k], list):
                inner_list = []
                for it in data[k]:
                    if isinstance(it, str):
                        inner_list.append(it)
                    else:
                        inner_list.append(json.loads(str(it)))
                data[k] = inner_list
            elif isinstance(data[k], datetime) or isinstance(data[k], date):
                data[k] = str(data[k])
            elif isinstance(data[k], Decimal):
                data[k] = float(data[k])
        return data

    def build(self):
        data = copy(self.__dict__)
        if self.__update_items is not None:
            data.update(self.__update_items)
        [data.pop(k, None) for k in self.__remove_keys]
        return data


class MutatorMode(Enum):
    DESERIALIZATION = "D"
    SERIALIZATION = "S"
    ALL = "*"


class FieldMutator(ABC):
    @abstractmethod
    def exec(self, value: Any) -> Any:
        ...


class ZMutator(FieldMutator):
    def __init__(self, mode: MutatorMode,
                 raise_err: bool = True,
                 error_to_raise: Optional[ZHttpError] = None,
                 error_msg: str = None,
                 cause_msg: str = None,
                 action=None,
                 predicate=None):
        self.mode = mode
        self.error_to_raise = error_to_raise
        self.error_msg = error_msg
        self.cause_msg = "Mutator execution failed" if cause_msg is None else cause_msg
        self.action = action
        self.raise_err = raise_err
        self.predicate = if_null_get(predicate, lambda x: True)
        self.can_run = True

    @classmethod
    def with_serialize(cls, raise_err: bool = True, error_msg: str = None, cause_msg: str = None,
                       action=None, error_to_raise: Optional[ZHttpError] = None, predicate=None):
        """
        Create mutator with serialization mode
        @param predicate:
        @param error_to_raise:
        @param raise_err:
        @param error_msg:
        @param cause_msg:
        @param action:
        @return:
        """
        return cls(MutatorMode.SERIALIZATION, raise_err, error_to_raise, error_msg, cause_msg, action, predicate)

    @classmethod
    def with_deserialize(cls, raise_err: bool = True, error_msg: str = None, cause_msg: str = None,
                         action=None, error_to_raise: Optional[ZHttpError] = None, predicate=None):
        """
        Create mutator with deserialization mode
        @param predicate:
        @param error_to_raise:
        @param raise_err:
        @param error_msg:
        @param cause_msg:
        @param action:
        @return:
        """
        return cls(MutatorMode.DESERIALIZATION, raise_err, error_to_raise, error_msg, cause_msg, action, predicate)

    @classmethod
    def with_all(cls, raise_err: bool = True, error_msg: str = None, cause_msg: str = None,
                 action=None, error_to_raise: Optional[ZHttpError] = None, predicate=None):
        """
        Create mutator with deserialization/serialization mode
        @param predicate:
        @param error_to_raise:
        @param raise_err:
        @param error_msg:
        @param cause_msg:
        @param action:
        @return:
        """
        return cls(MutatorMode.ALL, raise_err, error_to_raise, error_msg, cause_msg, action, predicate)

    def exec(self, value: str) -> str:
        if self.action is not None:
            if self.predicate(value):
                return self.action(value)
        return value


# _raise_when_failed = True
#
#
# def mutators_settings(raise_when_failed: bool = True) -> None:
#     global _raise_when_failed
#     _raise_when_failed = raise_when_failed


class Str(Field):
    """A string field.

    :param kwargs: The same keyword arguments that :class:`Field` receives.
    """

    #: Default error messages.
    default_error_messages = {
        "invalid": "Not a valid string.",
        "not_empty": "The string value can't be empty or null.",
        "invalid_utf8": "Not a valid utf-8 string.",
    }

    def _serialize(self, value, attr, obj, **kwargs) -> typing.Optional[str]:
        if value is None:
            return None
        return self.__apply_str_mappers(
            utils.ensure_text_type(value), MutatorMode.SERIALIZATION
        )

    def _get_mutators(self, mode: MutatorMode) -> List[ZMutator]:
        if 'mutators' in self.metadata:
            return list(filter(lambda m: m.mode == mode, self.metadata['mutators']))
        return []

    def __apply_str_mappers(self, value: str, mode: MutatorMode) -> str:
        mutators = self._get_mutators(mode)
        mutable_value: str = value
        for mutator in mutators:
            try:
                mutable_value = mutator.exec(mutable_value)
            except Exception as e:
                if mutator.raise_err is True:
                    if mutator.error_to_raise is not None:
                        mutator.error_to_raise.internal_exception = e
                        raise mutator.error_to_raise
                    raise BadRequest(message=mutator.error_msg, reason=mutator.cause_msg, parent_ex=e)
        return mutable_value

    def _deserialize(self, value, attr, data, **kwargs) -> typing.Any:

        if not isinstance(value, (str, bytes)):
            raise self.make_error("invalid")

        if 'allow_empty' in self.metadata and self.metadata['allow_empty'] is False:
            if not value.strip():
                raise self.make_error("not_empty")

        try:
            return self.__apply_str_mappers(
                utils.ensure_text_type(value), MutatorMode.DESERIALIZATION
            )
        except UnicodeDecodeError as error:
            raise self.make_error("invalid_utf8") from error


_T = typing.TypeVar("_T")


class Num(Field):
    """Base class for number fields.

    :param bool as_string: If `True`, format the serialized value as a string.
    :param kwargs: The same keyword arguments that :class:`Field` receives.
    """

    num_type = float  # type: typing.Type

    #: Default error messages.
    default_error_messages = {
        "invalid": "Not a valid number.",
        "too_large": "Number too large.",
    }

    def __init__(self, *, as_string: bool = False, **kwargs):
        self.as_string = as_string
        super().__init__(**kwargs)

    def _format_num(self, value) -> typing.Any:
        """Return the number value for value, given this field's `num_type`."""
        return self.num_type(value)

    def _validated(self, value) -> typing.Optional[_T]:
        """Format the value or raise a :exc:`ValidationError` if an error occurs."""
        if value is None:
            return None
        # (value is True or value is False) is ~5x faster than isinstance(value, bool)
        if value is True or value is False:
            raise self.make_error("invalid", input=value)
        try:
            return self._format_num(value)
        except (TypeError, ValueError) as error:
            raise self.make_error("invalid", input=value) from error
        except OverflowError as error:
            raise self.make_error("too_large", input=value) from error

    def _to_string(self, value) -> str:
        return str(value)

    def _serialize(
            self, value, attr, obj, **kwargs
    ) -> typing.Optional[typing.Union[str, _T]]:
        """Return a string if `self.as_string=True`, otherwise return this field's `num_type`."""
        print("hola serialize")
        if value is None:
            return None
        ret = self._format_num(value)  # type: _T
        return self._to_string(ret) if self.as_string else ret

    def _deserialize(self, value, attr, data, **kwargs) -> typing.Optional[_T]:
        print("hola des")
        return self._validated(value)
